
#include "mat.h"
#include "mex.h"
#include <iostream>
#include <ctime>
#include <cstdlib>
#include <fstream>
#include <stdexcept>
#include <ctime>
#include <Eigen/Dense>
#define LEARNING 1234
#define TESTING 4321
#define MIXING 5678
#define PULSE 1978
#define SPONTANEOUS 8654

#define dt 1.0

#define BASEALTD (14e-5 * 1.0 * 1.0)
#define RANDALTD 0.0
#define ALTP (8e-5  * .006  * 1.0 ) 
#define MINV -80.0
#define TAUVLONGTRACE 20000  

#define NBNEUR 100 

#define LATCONNMULTINIT 12.0
#define WFFINITMAX .2
#define WFFINITMIN 0.0
#define WPENSCALE .33
#define MAXW 50.0 
#define VSTIM 1.0

#define TIMEZEROINPUT 100
#define NBPATTERNSLEARNING 1000000
#define NBPATTERNSTESTING 3000 
#define NBPATTERNSPULSE 50 
#define PRESTIMELEARNING 350 // in ms
#define PRESTIMETESTING 350 // in ms
#define PRESTIMEMIXING 350 // in ms
#define PRESTIMEPULSE 350
#define PRESTIMESPONT 1000
#define PULSESTART 1 
//#define NBPRESPERPATTERNLEARNING 30 
#define NBPRESPERPATTERNTESTING 1 
#define NBMIXES 30

#define PATCHSIZE 17
#define FFRFSIZE (2 * PATCHSIZE * PATCHSIZE)

// Inhibition parameters
#define TAUINHIB 10 // in ms
#define ALPHAINHIB .75 // .6

#define NEGNOISERATE 1.0 // in KHz (expected number of thousands of VSTIM received per second through noise)
#define POSNOISERATE 2.5 // in KHz (expected number of thousands of VSTIM received per second through noise)

#define A 4
#define B .0805
#define Isp 400.0
#define TAUZ 40.0
#define TAUADAP 144.0
#define TAUVTHRESH 50.0
#define C 281.0
#define Gleak 30.0
#define Eleak -70.6
#define DELTAT 2.0  // in mV
#define VTMAX -30.4
#define VTREST -50.4
#define VPEAK 20  // Also in mV 
#define VRESET Eleak


#define THETAVLONGTRACE  -45.3 // -45.3 //MINV // Eleak // VTMAX

#define MAXDELAYDT 3
#define NBSPIKINGSTEPS 1 // (3.0 / dt) // Number of steps that a spike should take - i.e. spiking time (in ms) / dt.
#define REFRACTIME  0 // (dt-.001)
#define THETAVPOS -45.3
#define THETAVNEG Eleak
#define TAUXPLAST 15.0 // all 'tau' constants are in ms
#define TAUVNEG 10.0 
#define TAUVPOS 7.0
#define VREF2 50 // 70  // in mV^2

#define NBNOISESTEPS 73333


using namespace Eigen;
using namespace std;




MatrixXd poissonMatrix(const MatrixXd& lambd);
MatrixXd poissonMatrix2(const MatrixXd& lambd);
int poissonScalar(const double lambd);
void saveWeights(MatrixXd& wgt, const char *fn);
void readWeights(MatrixXd& wgt, const char *fn);

int main(int argc, char* argv[])
{
    //srand(8);
    clock_t tic;
    MATFile *pmat;
    mxArray *arr; double* imagedata;
    int PHASE;
    int STIM1, STIM2;
    int PULSETIME = 10;
    int NOLAT = 0;
    int NOINH = 0;
    int NONOISE = 0;
    double LATCONNMULT = LATCONNMULTINIT;
    int NBLASTRESPS =  50;
    double INPUTMULT = -1.0; // To be modified!
    if (argc > 1 && strcmp(argv[1], "test") == 0) PHASE = TESTING;
    else if (argc > 1 && strcmp(argv[1], "learn") == 0) PHASE = LEARNING;
    else if (argc > 1 && strcmp(argv[1], "mix") == 0) PHASE = MIXING;
    else if (argc > 1 && strcmp(argv[1], "pulse") == 0) PHASE = PULSE;
    else if (argc > 1 && strcmp(argv[1], "spont") == 0) PHASE = SPONTANEOUS;
    else { cerr << endl << "Error: You must provide at least one argument - 'learn', 'mix', 'pulse' or 'test'." << endl; if (argc >1) { cout<< "You provided argument '" << argv[1] << "'" << endl;}  return -1; }

    MatrixXd w = MatrixXd::Zero(NBNEUR, NBNEUR);
    MatrixXd wff = MatrixXd::Zero(NBNEUR, FFRFSIZE); 
    int NBPATTERNS, PRESTIME, NBPRES, NBSTEPSPERPRES, NBSTEPS, NBSTEPSVLONGTRACE;
    
            for (int nn=2; nn < argc; nn++){
                if (std::string(argv[nn]).compare("nonoise") == 0) {
                    NONOISE = 1;
                    cout << "No noise!" << endl;
                }
                if (std::string(argv[nn]).compare("noinh") == 0) {
                    NOINH = 1;
                    cout << "No inhibition!" << endl;
                }
                if (std::string(argv[nn]).compare("nolat") == 0) {
                    NOLAT = 1;
                    cout << "No lateral connections!" << endl;
                }
                if (std::string(argv[nn]).compare("latconn") == 0) {
                    LATCONNMULT = std::stod(argv[nn+1]);
                }
                if (std::string(argv[nn]).compare("pulsetime") == 0) {
                    PULSETIME = std::stoi(argv[nn+1]);
                }
            }

    if (PHASE == LEARNING)
    {
        NBPATTERNS = NBPATTERNSLEARNING; PRESTIME = PRESTIMELEARNING; NBPRES = NBPATTERNS; //* NBPRESPERPATTERNLEARNING;
        w =  MatrixXd::Zero(NBNEUR, NBNEUR); //MatrixXd::Random(NBNEUR, NBNEUR).cwiseAbs();
        wff =  (WFFINITMIN + (WFFINITMAX-WFFINITMIN) * MatrixXd::Random(NBNEUR, FFRFSIZE).cwiseAbs().array()).cwiseMin(MAXW) ; //MatrixXd::Random(NBNEUR, NBNEUR).cwiseAbs();
    }
    else if (PHASE == SPONTANEOUS)
    {
        NBPATTERNS = NBPATTERNSPULSE; PRESTIME = PRESTIMESPONT; NBPRES = NBPATTERNS; //* NBPRESPERPATTERNTESTING;
        NBLASTRESPS = NBPATTERNSPULSE;
        readWeights(w, "w.dat");
        readWeights(wff, "wff.dat");
        cout << "Spontaneous activity - no stimulus !" << endl;
    }
    else if (PHASE == PULSE)
    {
        NBPATTERNS = NBPATTERNSPULSE; PRESTIME = PRESTIMEPULSE; NBPRES = NBPATTERNS; //* NBPRESPERPATTERNTESTING;
        if (argc < 3) { cerr << endl << "Error: When using 'pulse', you must provide the number of the stimulus you want to pulse." << endl; return -1; }
        STIM1 = std::stoi(argv[2]);
        NBLASTRESPS = NBPATTERNSPULSE;
        cout << "Stim1: " << STIM1 << endl;
        readWeights(w, "w.dat");
        readWeights(wff, "wff.dat");
        cout << "Pulse input time: " << PULSETIME << " ms" << endl;
    }
    else if (PHASE == TESTING)
    {
        NBPATTERNS = NBPATTERNSTESTING; PRESTIME = PRESTIMETESTING; NBPRES = NBPATTERNS; //* NBPRESPERPATTERNTESTING;
        readWeights(w, "w.dat");
        readWeights(wff, "wff.dat");
    }
    else if (PHASE == MIXING)
    {
        NBPATTERNS = 2; PRESTIME = PRESTIMEMIXING ; NBPRES = NBMIXES * 3; //* NBPRESPERPATTERNTESTING;
        readWeights(w, "w.dat");
        readWeights(wff, "wff.dat");
        if (argc < 4) { cerr << endl << "Error: When using 'mix', you must provide the numbers of the 2 stimuli you want to mix." << endl; return -1; }
        STIM1 = std::stoi(argv[2]);
        STIM2 = std::stoi(argv[3]);
        cout << "Stim1, Stim2: " << STIM1 << ", " << STIM2 <<endl;
        if (argc == 5) 
            if (std::string(argv[4]).compare("nolat") == 0)
            {
                NOLAT = 1;
                cout << "No lateral connections!" << endl;
            }
        // wff *= 2; w *= 2;
    }
    else { cerr << "Which phase?\n"; return -1;}
    cout << "Lat. conn. mult. : " << LATCONNMULT << endl;
    NBSTEPSPERPRES = (int)(PRESTIME / dt);
    NBSTEPS = NBSTEPSPERPRES * NBPRES;
    // Note that keeping the vlongtrace window short seems to help with epilepsies
    NBSTEPSVLONGTRACE = 6 * NBSTEPSPERPRES;

    cout << NBSTEPSPERPRES << endl;
    MatrixXi lastnresps = MatrixXi::Zero(NBNEUR, NBLASTRESPS * NBSTEPSPERPRES);

    cout << "Reading input data...." << endl;
    //if (!(pmat = matOpen("../patches.mat", "r"))) { cout << " ERROR: Couldn't open frames file!\n" << endl; return -1;} 

    // WE USE THIS ONE BUT THEN WE SCALE BY MAX IN LGN RATE PREPARATION!!
    if (!(pmat = matOpen("../patchesCenteredScaledBySumTo70EachFrameRotated.mat", "r"))) { cout << " ERROR: Couldn't open frames file!\n" << endl; return -1;}  //INPUTMULT = 200.0; // When normalizing by sum, you need an inputmult...

    //if (!(pmat = matOpen("../patchesCenteredScaledTo70EachFrame.mat", "r"))) { cout << " ERROR: Couldn't open frames file!\n" << endl; return -1;} INPUTMULT = 1.0; //2.00;
    if (!(arr = matGetVariable(pmat, "patches"))) { cout << " ERROR: couldn't access variable!\n" << endl; return -1;}
    int nbdim = mxGetNumberOfDimensions(arr); for (int nn=0; nn < nbdim; nn++) cout << "Size of array 'patches' in dimension " << nn << ": " << mxGetDimensions(arr)[nn] << endl;
    int nbpatchesinfile = (int)(mxGetDimensions(arr)[0] / (PATCHSIZE * PATCHSIZE)) - 1;
    int totaldatasize = mxGetNumberOfElements(arr);
    imagedata = (double *) mxGetData(arr);
    matClose(pmat);

    
    /*cout << "Preparing stimulus input...." << endl;
    // First, we create the different input patterns, expressed as mean rate of inputs for each neuron, in KHz (i.e. number inputs per ms)
    MatrixXd patterns = MatrixXd::Zero(NBNEUR, NBPATTERNS);
    for (int nump=0; nump < NBPATTERNS; nump++)
    {
        patterns.col(nump) = (5 * VectorXd::Random(NBNEUR).cwiseAbs() - VectorXd::Constant(NBNEUR, 4)).cwiseMax(0);
        //patterns.col(nump) *= 8.0 * (36.0 / NBNEUR) ;
        patterns.col(nump) *= INPUTMULT * (36.0 / NBNEUR) ;
    }
    cout << patterns << endl << endl;*/


    // The noise excitatory input is a Poisson process (separate for each cell) with a constant rate (in KHz / per ms)
    MatrixXd negnoisein = - poissonMatrix(dt * MatrixXd::Constant(NBNEUR, NBNOISESTEPS, NEGNOISERATE)) * VSTIM;
    MatrixXd posnoisein = poissonMatrix(dt * MatrixXd::Constant(NBNEUR, NBNOISESTEPS, POSNOISERATE)) * VSTIM;
    if (NONOISE){
        posnoisein.setZero();
        negnoisein.setZero();
    }

    VectorXd v =  VectorXd::Constant(NBNEUR, -70.5); // VectorXd::Zero(NBNEUR); // -70.5 is approximately the resting potential of the Izhikevich neurons, as it is of the AdEx neurons used in Clopath's experiments


    //MatrixXd vs = MatrixXd::Zero(NBNEUR, NBSTEPS);
    //MatrixXi spikes = MatrixXi::Zero(NBNEUR, NBSTEPS);
    VectorXi firings = VectorXi::Zero(NBNEUR);
    VectorXi firingsprev = VectorXi::Zero(NBNEUR);
    VectorXi firingsprevprev = VectorXi::Zero(NBNEUR);
    VectorXi firingsprevprevprev = VectorXi::Zero(NBNEUR);
    VectorXi firingsprevprevprevprev = VectorXi::Zero(NBNEUR);
    MatrixXi outgoingspikes = MatrixXi::Zero(NBNEUR, MAXDELAYDT);
    VectorXd Iff = VectorXd::Zero(NBNEUR);
    VectorXd Ilat = VectorXd::Zero(NBNEUR);
    VectorXd I;
    VectorXd xplast_ff = VectorXd::Zero(FFRFSIZE);
    VectorXd xplast_lat = VectorXd::Zero(NBNEUR);
    VectorXd vneg = v;
    VectorXd vpos = v;
    VectorXd vprev = v;
    VectorXd vprevprev = v;

    // Correct 
    VectorXd vlongtrace = (v.array() - THETAVLONGTRACE).cwiseMax(0);

    // Wrong:
    //VectorXd vlongtrace = v;

    VectorXd InhibVect = VectorXd::Zero(NBNEUR);
    VectorXi ZeroV = VectorXi::Zero(NBNEUR); VectorXi OneV = VectorXi::Constant(NBNEUR, 1); 
    VectorXd ZeroLGN = VectorXd::Zero(FFRFSIZE); VectorXd OneLGN = VectorXd::Constant(FFRFSIZE, 1.0); 
    VectorXd z = VectorXd::Zero(NBNEUR);
    VectorXd wadap = VectorXd::Zero(NBNEUR);
    VectorXd vthresh = VectorXd::Constant(NBNEUR, VTREST);
    VectorXd refractime  = VectorXd::Zero(NBNEUR);
    VectorXi isspiking = VectorXi::Zero(NBNEUR);
    VectorXd EachNeurLTD = VectorXd::Zero(NBNEUR);
    VectorXd EachNeurLTP = VectorXd::Zero(NBNEUR);
    MatrixXd vnegvals = vlongtrace.rowwise().replicate(NBSTEPSVLONGTRACE); // MatrixXd::Zero(NBNEUR, NBSTEPSVLONGTRACE); Doesn't work with the mean algorithm!
    double inhibrate = .5; 


    double ALTDS[NBNEUR]; for (int nn=0; nn < NBNEUR; nn++) ALTDS[nn] = BASEALTD + RANDALTD*( (double)rand() / (double)RAND_MAX );

    VectorXd lgnrates = VectorXd::Zero(FFRFSIZE);
    VectorXd lgnratesS1 = VectorXd::Zero(FFRFSIZE);
    VectorXd lgnratesS2 = VectorXd::Zero(FFRFSIZE);
    VectorXd lgnfirings = VectorXd::Zero(FFRFSIZE);

    //VectorXd inhibs = VectorXd::Zero(NBSTEPS);
    VectorXd sumwff = VectorXd::Zero(NBPRES);
    VectorXd sumw = VectorXd::Zero(NBPRES);
    MatrixXi resps = MatrixXi::Zero(NBNEUR, NBPRES);
    /*MatrixXd lgninputs = MatrixXd::Zero(FFRFSIZE, NBPRES);
    MatrixXd meanvlongtrace = MatrixXd::Zero(NBNEUR, NBPRES);
    MatrixXd meanvneg = MatrixXd::Zero(NBNEUR, NBPRES);
    MatrixXd meanvpos = MatrixXd::Zero(NBNEUR, NBPRES);
    MatrixXd meanENLTP = MatrixXd::Zero(NBNEUR, NBPRES);
    MatrixXd meanENLTD = MatrixXd::Zero(NBNEUR, NBPRES);*/
    
    double mixvals[NBMIXES]; for (int nn=0; nn < NBMIXES; nn++) mixvals[nn] = (double)nn / (double)(NBMIXES - 1); // NBMIXES values equally spaced from 0 to 1 inclusive.
    //STIM1 = 940000 - 1002 + 310; //779; 
    //STIM2 = 940000 - 1002 + 630; //740;


    fstream myfile;
    tic = clock();
    int numstep = 0;
    for (int numpres=0; numpres < NBPRES; numpres++)
    {
        for (int nn=0; nn < (NBLASTRESPS -1) * NBSTEPSPERPRES; nn++)
            lastnresps.col(nn) = lastnresps.col(nn + NBSTEPSPERPRES);
     

        int posindata = ((numpres % nbpatchesinfile) * FFRFSIZE / 2 ); 
        if (PHASE == PULSE)
            posindata = ((STIM1 % nbpatchesinfile) * FFRFSIZE / 2 ); 
        if (posindata >= totaldatasize - FFRFSIZE / 2) { cerr << "Error: tried to read beyond data end.\n"; return -1; }

        //cout << posindata << endl;


        // Extracting the image data for this frame presentation, and preparing the LGN output rates
        
        for (int nn=0; nn < FFRFSIZE / 2; nn++)
        {
            lgnrates(nn) = log(1.0 + (imagedata[posindata+nn] > 0 ? imagedata[posindata+nn] : 0));
            lgnrates(nn + FFRFSIZE / 2) = log(1.0 + (imagedata[posindata+nn] < 0 ? -imagedata[posindata+nn] : 0));
            //lgnrates(nn) = (imagedata[posindata+nn] > 0 ? imagedata[posindata+nn] : 0);
            //lgnrates(nn + FFRFSIZE / 2) = (imagedata[posindata+nn] < 0 ? -imagedata[posindata+nn] : 0);
        }
        lgnrates /= lgnrates.maxCoeff(); // Scale by max!!
   
        /*
    % To find the "good" stimuli: after running analyzewlat.m :
    % ss=sum(sr(oRP,:), 2)'; % look at the positions in the response pattern matrix that correspond to a given cluster and have high ss (total firing) value; then look at the corresponding position in oRP, which gives you the number of the actual presentation (with the 200K-1002 shift).


    */



        if (PHASE == MIXING)
        {
            int posindata1 = ((STIM1 % nbpatchesinfile) * FFRFSIZE / 2 ); if (posindata1 >= totaldatasize - FFRFSIZE / 2) { cerr << "Error: tried to read beyond data end.\n"; return -1; }
            int posindata2 = ((STIM2 % nbpatchesinfile) * FFRFSIZE / 2 ); if (posindata2 >= totaldatasize - FFRFSIZE / 2) { cerr << "Error: tried to read beyond data end.\n"; return -1; }

            double mixval1 = mixvals[numpres % NBMIXES]; double mixval2 = 1.0-mixval1; double mixedinput = 0;
            if ((numpres / NBMIXES) == 1 ) mixval2 = 0;
            if ((numpres / NBMIXES) == 2 ) mixval1 = 0;
            
            for (int nn=0; nn < FFRFSIZE / 2; nn++)
            {
                lgnratesS1(nn) = log(1.0 + (imagedata[posindata1+nn] > 0 ? imagedata[posindata1+nn] : 0));  lgnratesS1(nn + FFRFSIZE / 2) = log(1.0 + (imagedata[posindata1+nn] < 0 ? -imagedata[posindata1+nn] : 0));
                lgnratesS2(nn) = log(1.0 + (imagedata[posindata2+nn] > 0 ? imagedata[posindata2+nn] : 0));  lgnratesS2(nn + FFRFSIZE / 2) = log(1.0 + (imagedata[posindata2+nn] < 0 ? -imagedata[posindata2+nn] : 0));
                //lgnratesS1(nn) = ( (imagedata[posindata1+nn] > 0 ? imagedata[posindata1+nn] : 0));  lgnratesS1(nn + FFRFSIZE / 2) = ( (imagedata[posindata1+nn] < 0 ? -imagedata[posindata1+nn] : 0));
                //lgnratesS2(nn) = ( (imagedata[posindata2+nn] > 0 ? imagedata[posindata2+nn] : 0));  lgnratesS2(nn + FFRFSIZE / 2) = ( (imagedata[posindata2+nn] < 0 ? -imagedata[posindata2+nn] : 0));
            }
            lgnratesS1 /= lgnratesS1.maxCoeff(); // Scale by max!!
            lgnratesS2 /= lgnratesS2.maxCoeff(); // Scale by max!!
            
            for (int nn=0; nn < FFRFSIZE ; nn++)
                lgnrates(nn) = mixval1 * lgnratesS1(nn) + mixval2 * lgnratesS2(nn);
            
        }


        INPUTMULT = 150.0;
        //lgnrates /= lgnrates.norm(); // Scale by norm!!
        //INPUTMULT = 150.0 * 5.4 / 1.16;
        INPUTMULT *= 2.0;



        //INPUTMULT /= 5.0; // /=10.0;



        lgnrates *= INPUTMULT; // We put inputmult here to ensure that it is reflected in the actual number of incoming spikes




        //lgnrates.array() -= lgnrates.mean();



        lgnrates *= (dt/1000.0);  // LGN rates from the pattern file are expressed in Hz. We want it in rate per dt, and dt itself is expressed in ms.

        // Simple stimuli
        //for (int nn=0; nn < FFRFSIZE; nn++)    
        //    lgnrates(nn) = ((nn + (numpres % 3)) % 3 == 0 ? .03 :  0.0); // in KHz
        
        //cout << "LGN rates: " << endl << lgnrates.transpose() << endl;



        v.fill(Eleak);
        firings.setZero();
        firingsprev.setZero();
        firingsprevprev.setZero();
        firingsprevprevprev.setZero();
        firingsprevprevprevprev.setZero();

        // Stimulus presentation
        for (int numstepthispres=0; numstepthispres < NBSTEPSPERPRES; numstepthispres++)
        {
            //if (numstep % 10000 == 0) { cout << numstep << " / " << NBSTEPS << ". Inhib rate (per timestep):"  << inhibrate << endl; }

            if (((PHASE == PULSE) && (numstepthispres >= (double)(PULSESTART)/dt) && (numstepthispres < (double)(PULSESTART + PULSETIME)/dt)) // In the PULSE case, inputs only fire for a short period of time
                    || ((PHASE != PULSE) && (numstepthispres < NBSTEPSPERPRES - ((double)TIMEZEROINPUT / dt))))   // Otherwise, inputs only fire until the 'relaxation' period at the end of each presentation
                for (int nn=0; nn < FFRFSIZE; nn++)
                    lgnfirings(nn) = (rand() / (double)RAND_MAX < abs(lgnrates(nn)) ? 1.0 : 0.0); // Note that this may go non-poisson if the specified lgnrates are too high (i.e. not << 1.0)
            else
                lgnfirings.setZero();

            if (PHASE == SPONTANEOUS)
                lgnfirings.setZero();

 
            lgnfirings = (lgnrates.array()<0).select(-lgnfirings, lgnfirings);

            Iff =  wff * lgnfirings * VSTIM;





            //Iff = Iff / 2.0;





            inhibrate += (dt / TAUINHIB) * (ALPHAINHIB * firings.sum() * (36.0 / 5.0) - inhibrate);
            //inhibrate += (dt / TAUINHIB) * (ALPHAINHIB * firingsprevprevprevprev.sum() * (36.0 / 5.0) - inhibrate);
            
            //inhibs(numstep) = inhibrate;
            InhibVect = poissonMatrix2(VectorXd::Constant(NBNEUR, inhibrate)) ; 

            //Ilat = VSTIM * LATCONNMULT * (36.0 / 100.0) * w * firings.cast<double>();
            //Ilat = VSTIM * LATCONNMULT * (36.0 / 100.0) * w * firingsprevprev.cast<double>();
            //Ilat = VSTIM * LATCONNMULT * (36.0 / 100.0) * w * ((VectorXd::Random(NBNEUR)).array() > 0).select(firingsprevprev.cast<double>(), firingsprev.cast<double>());

            VectorXi LatInput = VectorXi::Zero(NBNEUR);
                // delays
            LatInput = outgoingspikes.col(numstep % MAXDELAYDT);
            outgoingspikes.col(numstep % MAXDELAYDT).setZero();

            if (1 == 0)
                for (int nn=0; nn < NBNEUR; nn++)
                {
                    double r = (double)rand() / (double)RAND_MAX; 
                    /*if (r < .50)
                        LatInput(nn) = firingsprev(nn);
                    else 
                        LatInput(nn) = firings(nn);*/
                    /*if (r < .20)
                        LatInput(nn) = firingsprev(nn);
                    else if (r < .4)
                        LatInput(nn) = firingsprevprev(nn);
                    else if (r < .6)
                        LatInput(nn) = firingsprevprevprev(nn);
                    else if (r < .8)
                        LatInput(nn) = firings(nn);
                    else 
                        LatInput(nn) = firingsprevprevprevprev(nn);*/
                    //LatInput(nn) = firings(nn);
                }


            Ilat = VSTIM * LATCONNMULT * (36.0 / 100.0) * w * LatInput.cast<double>();

            if (NOLAT)
                Ilat.setZero();
            if (NOINH)
                InhibVect.setZero();
            
            I = Iff + Ilat + posnoisein.col(numstep % NBNOISESTEPS) + negnoisein.col(numstep % NBNOISESTEPS) - InhibVect;

            vprev = v;
            vprevprev = vprev;

            // AdEx version:
            for (int nn=0; nn < NBNEUR; nn++)
                v(nn) += (dt/C) * (-Gleak * (v(nn)-Eleak) + Gleak * DELTAT * exp( (v(nn)-vthresh(nn)) / DELTAT ) + z(nn) - wadap(nn) ) + I(nn);
                //v(nn) += (dt/C) * (-Gleak * (v(nn)-Eleak) + Gleak * DELTAT * exp( (v(nn)-vthresh(nn)) / DELTAT ) + z(nn) - wadap(nn)  + I(nn));  // The input current is also included in the diff. eq.


            //v = (refractime.array() > 0).select(Eleak, v); // NO REFRACTORY TIME!
           
            v = (isspiking.array() > 0).select(VPEAK-.001, v); // Currently-spiking neurons are clamped at VPEAK.
            v = (isspiking.array() == 1).select(VRESET, v); //  Neurons that have finished their spiking are set to VRESET.
            
            z = (isspiking.array() == 1).select(Isp, z);
            vthresh = (isspiking.array() == 1).select(VTMAX, vthresh);
            wadap = (isspiking.array() == 1).select(wadap.array() + B, wadap.array());
            
            isspiking = (isspiking.array() - 1).cwiseMax(0);
            v = v.cwiseMax(MINV);
            refractime = (refractime.array() - dt).cwiseMax(0);
            

            // "correct" version: Firing neurons are crested / clamped at VPEAK, will be reset to VRESET  after the spiking time has elapsed.
            firingsprevprevprevprev = firingsprevprevprev;
            firingsprevprevprev = firingsprevprev;
            firingsprevprev = firingsprev;
            firingsprev = firings;
            firings = (v.array() > VPEAK).select(OneV, ZeroV);
            v = (firings.array() > 0).select(VPEAK, v);
            refractime = (firings.array() > 0).select(REFRACTIME, refractime);
            isspiking = (firings.array() > 0).select(NBSPIKINGSTEPS, isspiking);


            // For each spike, the delay is a number between 1 and MAXDELAYDT inclusive.
            for (int nn=0; nn < NBNEUR; nn++)
                outgoingspikes(nn, (numstep +  (int)ceil(MAXDELAYDT * (double)rand() / (double)RAND_MAX)) % MAXDELAYDT ) += firings(nn);


            // "Wrong" version: firing if above threshold, immediately reset at Vreset.
            //firings = (v.array() > vthresh.array()).select(OneV, ZeroV);
            //v = (firings.array() > 0).select(VRESET, v);


            
            // AdEx version:

            //wadap = (isspiking.array() > 0).select(wadap.array(), wadap.array() + (dt / TAUADAP) * (A * (v.array() - Eleak) - wadap.array())); // clopathlike (while spiking, don't modify wadap.
            wadap =  wadap.array() + (dt / TAUADAP) * (A * (v.array() - Eleak) - wadap.array());
            z = z + (dt / TAUZ) * -1.0 * z;
            vthresh = vthresh.array() + (dt / TAUVTHRESH) * (-1.0 * vthresh.array() + VTREST);
            
            // Wrong - using the raw v rather than "depolarization" v-vleak (or v-vthresh)
            //vlongtrace = vlongtrace + (dt / TAUVLONGTRACE) * (v - vlongtrace);

            // Correct: using depolarization (or more precisely depolarization above THETAVLONGTRACE))
            vlongtrace += (dt / TAUVLONGTRACE) * ((vprevprev.array() - THETAVLONGTRACE).cwiseMax(0).matrix() - vlongtrace);
            vlongtrace = vlongtrace.cwiseMax(0); // Just in case.
            
            // This is also wrong - the dt/tau should not apply to the increments (firings / lgnfirings). However that should only be a strict constant multiplication, which could be included into the ALTP/ALTP constants.
            /*
            xplast_lat += (dt / TAUXPLAST) * (firings.cast<double>() - xplast_lat);
            xplast_ff += (dt / TAUXPLAST) * (lgnfirings - xplast_ff);
            vneg += (dt / TAUVNEG) * (v - vneg);
            vpos += (dt / TAUVPOS) * (v - vpos);*/

            // "Correct" version:
            //xplast_lat = xplast_lat + firings.cast<double>() - (dt / TAUXPLAST) * xplast_lat;
            //xplast_ff = xplast_ff + lgnfirings - (dt / TAUXPLAST) *  xplast_ff;

            //Clopath-like version:
            xplast_lat = xplast_lat + firings.cast<double>() / TAUXPLAST - (dt / TAUXPLAST) * xplast_lat;
            xplast_ff = xplast_ff + lgnfirings / TAUXPLAST - (dt / TAUXPLAST) *  xplast_ff;

            vneg = vneg + (dt / TAUVNEG) * (vprevprev - vneg);
            vpos = vpos + (dt / TAUVPOS) * (vprevprev - vpos);
            

            if ((PHASE == LEARNING)  && (numpres >= 401) )
            {

                
                // For each neuron, we compute the quantities by which any synapse reaching this given neuron should be modified, if the synapse's firing / recent activity (xplast) commands modification.
                for (int nn=0; nn <  NBNEUR; nn++)
                    EachNeurLTD(nn) =  dt * (-ALTDS[nn] / VREF2) * vlongtrace(nn) * vlongtrace(nn) * ((vneg(nn) - THETAVNEG) <0 ? 0 : (vneg(nn) -THETAVNEG));
                    //EachNeurLTD(nn) =  dt * (-ALTDS[nn] / VREF2) * vlongtrace(nn) * vlongtrace(nn) * vlongtrace(nn) * (vneg(nn) - THETAVNEG <0 ? 0 : vneg(nn) -THETAVNEG); // Cube
                for (int nn=0; nn < NBNEUR; nn++)
                    EachNeurLTP(nn) =  dt * ALTP  * ((vpos(nn) - THETAVNEG)<0 ? 0 : (vpos(nn) - THETAVNEG) ) * ((v(nn) - THETAVPOS) < 0 ? 0 : (v(nn) - THETAVPOS));
            //EachNeurLTD = dt * (-ALTD / VREF2) * vlongtrace.cwiseProduct(vlongtrace).array().cwiseProduct( (vneg.array() - THETAVNEG).cwiseMax(0) );
            //EachNeurLTP = dt * ALTP * (vpos.array() - THETAVNEG).cwiseMax(0).cwiseProduct( (v.array() - THETAVPOS).cwiseMax(0) );

                // Feedforward synapses, then lateral synapses.
                for (int syn=0; syn < FFRFSIZE; syn++)
                        for (int nn=0; nn < NBNEUR; nn++)
                            wff(nn, syn) += xplast_ff(syn) * EachNeurLTP(nn);
                for (int syn=0; syn < FFRFSIZE; syn++)
                    if (lgnfirings(syn) > 1e-10)
                        for (int nn=0; nn < NBNEUR; nn++)
                                wff(nn, syn) +=  EachNeurLTD(nn) * (1.0 + wff(nn,syn) * WPENSCALE);
                for (int syn=0; syn < NBNEUR; syn++)
                        for (int nn=0; nn < NBNEUR; nn++)
                            w(nn, syn) += xplast_lat(syn) * EachNeurLTP(nn);
                for (int syn=0; syn < NBNEUR; syn++)
                    if (firings(syn) > 1e-10)
                        for (int nn=0; nn < NBNEUR; nn++)
                                w(nn, syn) +=  EachNeurLTD(nn) * (1.0 + w(nn,syn) * WPENSCALE);
                        
                w = w - w.cwiseProduct(MatrixXd::Identity(NBNEUR, NBNEUR)); // Diagonal lateral weights are 0!
                wff = wff.cwiseMax(0);
                w = w.cwiseMax(0);
                wff = wff.cwiseMin(MAXW);
                w = w.cwiseMin(MAXW);
            }
            //v = (firings.array() > 0).select(VRESET, v);

            //vs.col(numstep) = v;
            //spikes.col(numstep) = firings;
            resps.col(numpres) += firings;
            lastnresps.col((NBLASTRESPS - 1)*NBSTEPSPERPRES + numstepthispres) = firings;
            /*meanvlongtrace.col(numpres) += vlongtrace; 
            meanvneg.col(numpres) += vneg; 
            meanvpos.col(numpres) += vpos; 
            meanENLTP.col(numpres) += EachNeurLTP; 
            meanENLTD.col(numpres) += EachNeurLTD; */
        //    if (numstep > 201000) { toc = clock(); cout << endl << "Time spent, in seconds (excluding stimulus preparation): " << (double)(toc - tic) / CLOCKS_PER_SEC << endl; return 0; } 
            numstep++;
        }



        sumwff(numpres) = wff.sum();
        sumw(numpres) = w.sum();
        /*lgninputs.col(numpres) = lgnrates;
        meanvlongtrace.col(numpres) /= NBSTEPSPERPRES;
        meanvneg.col(numpres) /= NBSTEPSPERPRES;
        meanvpos.col(numpres) /= NBSTEPSPERPRES;
        meanENLTP.col(numpres) /= NBSTEPSPERPRES;
        meanENLTD.col(numpres) /= NBSTEPSPERPRES;*/
        if (numpres % 100 == 0)
        {
        cout << "Presentation " << numpres << " / " << NBPRES << endl; if (numpres == 200) cout << "TIME: " << (double)(clock() - tic) / (double)CLOCKS_PER_SEC <<endl;
            cout << "Total spikes for each neuron for this presentation: " << resps.col(numpres).transpose() << endl;
        cout << "Vlongtraces: " << vlongtrace.transpose() << endl;
        cout << " Max LGN rate (should be << 1.0): " << lgnrates.maxCoeff() << endl;
        }
        if ( ((numpres +1)%  10000 == 0) || (numpres+1 == NBPRES))
        {

            if (PHASE == TESTING){
                myfile.open("resps_test.txt", ios::trunc | ios::out);  myfile << endl << resps << endl; myfile.close();
            }
            if (PHASE == SPONTANEOUS){
                std::string nolatindicator ("");
                std::string noinhindicator ("");
                if (NOINH) noinhindicator = "_noinh";
                if (NOLAT) nolatindicator = "_nolat";
                myfile.open("resps_spont.txt", ios::trunc | ios::out);  myfile << endl << resps << endl; myfile.close();
                myfile.open("lastnresps_spont"+nolatindicator+noinhindicator+".txt", ios::trunc | ios::out);  myfile << endl << lastnresps << endl; myfile.close();
            }
            if (PHASE == PULSE){
                std::string nolatindicator ("");
                if (NOLAT) nolatindicator = "_nolat";
                myfile.open("lastnresps_pulse" + nolatindicator + ".txt", ios::trunc | ios::out);  myfile << endl << lastnresps << endl; myfile.close();
                myfile.open("resps_pulse.txt", ios::trunc | ios::out);  myfile << endl << resps << endl; myfile.close();
                myfile.open("resps_pulse_"+std::to_string((long long int)STIM1)+".txt", ios::trunc | ios::out);  myfile << endl << resps << endl; myfile.close();
            }
            if (PHASE == MIXING){
                std::string nolatindicator ("");
                if (NOLAT) nolatindicator = "_nolat";
                myfile.open("resps_mix" + nolatindicator + ".txt", ios::trunc | ios::out);  myfile << endl << resps << endl; myfile.close();
                //myfile.open("resps_mix.txt", ios::trunc | ios::out);  myfile << endl << resps << endl; myfile.close();
                myfile.open("resps_mix_"+std::to_string((long long int)STIM1)+"_"+std::to_string((long long int)STIM2) + nolatindicator + ".txt", ios::trunc | ios::out);  myfile << endl << resps << endl; myfile.close();
            }
            if (PHASE == LEARNING){
                cout << "(Saving temporary data ... )" << endl;

                myfile.open("w.txt", ios::trunc | ios::out);
                myfile << endl << w << endl;
                myfile.close();
                myfile.open("wff.txt", ios::trunc | ios::out);  myfile << endl << wff << endl; myfile.close();
                if ((numpres +1)%  10000 == 0) { char tmpstr[80]; sprintf(tmpstr, "%s%d%s", "wff_", numpres+1, ".txt"); myfile.open(tmpstr, ios::trunc | ios::out);  myfile << endl << wff << endl; myfile.close(); }
                myfile.open("resps.txt", ios::trunc | ios::out);
                myfile << endl << resps << endl;
                myfile.close();
                //myfile.open("patterns.txt", ios::trunc | ios::out);  myfile << endl << patterns << endl; myfile.close();
                /*myfile.open("lgninputs.txt", ios::trunc | ios::out); myfile << endl << lgninputs << endl; myfile.close();
                myfile.open("meanvneg.txt", ios::trunc | ios::out); myfile << endl << meanvneg << endl; myfile.close();
                myfile.open("meanvpos.txt", ios::trunc | ios::out); myfile << endl << meanvpos << endl; myfile.close();
                myfile.open("meanENLTD.txt", ios::trunc | ios::out); myfile << endl << meanENLTD << endl; myfile.close();
                myfile.open("meanENLTP.txt", ios::trunc | ios::out); myfile << endl << meanENLTP << endl; myfile.close();*/
                /*myfile.open("meanvlt.txt", ios::trunc | ios::out);
                myfile << endl << meanvlongtrace << endl;
                myfile.close();*/
                myfile.open("sumwff.txt", ios::trunc | ios::out); myfile << endl << sumwff << endl;
                myfile.open("sumwff.txt", ios::trunc | ios::out); myfile << endl << sumwff << endl;
                myfile.close();
                saveWeights(w, "w.dat");
                saveWeights(wff, "wff.dat");
            }
        }

    }

    // RESPS contains the total firing rate of all neurons for each input pattern presentation
    /*MatrixXd RESPS = MatrixXd::Zero(NBNEUR, NBPATTERNS * NBPRES);
    for (int numresp=0; numresp < NBPATTERNS*NBPRES; numresp++){
        RESPS.col(numresp) = spikes.block(0,numresp * BLOCKSIZE, NBNEUR, BLOCKSIZE).rowwise().sum() / (BLOCKSIZE * dt / 1000.0);
    }*/

    //cout << endl << "Weights:\n" << w << endl;
    //cout << endl << "Resps:\n" << resps << endl;
    /*
    myfile.open("spikes.txt", ios::trunc | ios::out);
    myfile << endl << spikes << endl;
    myfile.close();
    myfile.open("stimin.txt", ios::trunc | ios::out);
    myfile << endl << stimin << endl;
    myfile.close();
    myfile.open("allpatterns.txt", ios::trunc | ios::out);
    myfile << endl << allpatterns << endl;
    myfile.close();
    myfile.open("vs.txt", ios::trunc | ios::out);
    myfile << endl << endl<< vs << endl;
    myfile.close();
    */


}





/*
 *  Utility functions
 */

MatrixXd poissonMatrix2(const MatrixXd& lambd)
{
    MatrixXd k = MatrixXd::Zero(lambd.rows(),lambd.cols());
    for (int nr = 0; nr < lambd.rows(); nr++)
        for (int nc = 0; nc < lambd.cols(); nc++)
            k(nr, nc) = poissonScalar(lambd(nr, nc));
    return k;
}

MatrixXd poissonMatrix(const MatrixXd& lambd)
{
    //MatrixXd lambd = MatrixXd::Random(SIZ,SIZ).cwiseAbs();
    //MatrixXd lambd = MatrixXd::Random(SIZ,SIZ).cwiseMax(0);
    //MatrixXd lambd = MatrixXd::Constant(SIZ,SIZ, .5);

    MatrixXd L = (-1 * lambd).array().exp();
    MatrixXd k = MatrixXd::Zero(lambd.rows(),lambd.cols());
    MatrixXd p = MatrixXd::Constant(lambd.rows(),lambd.cols(),1.0);
    MatrixXd matselect = MatrixXd::Constant(lambd.rows(),lambd.cols(),1.0);

    while ((matselect.array() >0).any())
    {
        k = (matselect.array() > 0).select(k.array() + 1, k); // wherever p > L (after the first loop, otherwise everywhere), k += 1
        p = p.cwiseProduct(MatrixXd::Random(p.rows(),p.cols()).cwiseAbs());  // p = p * random[0,1]
        matselect = (p.array() > L.array()).select(matselect, -1.0);
    }

    k = k.array() - 1;
    return k;
}
/* 
 // Test code for poissonMatrix:
    double SIZ=19;
    srand(time(NULL));
    MatrixXd lbd;
    MatrixXd kout;
    double dd = 0;
    for (int nn = 0; nn < 10000; nn++)
    {
        lbd = MatrixXd::Random(SIZ,SIZ).cwiseMax(0);
        kout = poissonMatrix(lbd);
       dd += lbd.sum();
    }
    cout << endl << lbd << endl;
    cout << endl << kout << endl;
    cout << kout.mean() << endl;
    cout << dd << endl;
*/


int poissonScalar(const double lambd)
{
    double L = exp(-1 * lambd);
    int k =  0; double p = 1.0;
    do{
        k = k + 1;
        p = p * (double)rand() / (double)RAND_MAX;
    } 
    while(p > L);
    return (k-1);
}

void saveWeights(MatrixXd& wgt, const char* fname)
{
    double wdata[wgt.rows() * wgt.cols()];
    int idx=0;
    //cout << endl << "Saving weights..." << endl;
    for (int cc=0; cc < wgt.cols(); cc++)
        for (int rr=0; rr < wgt.rows(); rr++)
            wdata[idx++] = wgt(rr, cc);

    ofstream myfile(fname, ios::binary | ios::trunc);
    if (!myfile.write((char*) wdata, wgt.rows() * wgt.cols() * sizeof(double)))
        throw std::runtime_error("Error while saving matrix of weights.\n");
    myfile.close();
}

void readWeights(MatrixXd& wgt, const char* fname)
{
    double wdata[wgt.cols() * wgt.rows()];
    int idx=0;
    cout << endl << "Reading weights from file " << fname << endl;
    ifstream myfile(fname, ios::binary);
    if (!myfile.read((char*) wdata, wgt.cols() * wgt.rows() * sizeof(double)))
        throw std::runtime_error("Error while reading matrix of weights.\n");
    myfile.close();
    for (int cc=0; cc < wgt.cols() ; cc++)
        for (int rr=0; rr < wgt.rows(); rr++)
            wgt(rr, cc) = wdata[idx++];
    cout << "Done!" <<endl;
}

/* Previous, slow version of the plasticity:
            w = w.array() + dt * (-(ALTD / VREF2) * (vlongtrace.cwiseProduct(vlongtrace)).array().replicate(1, NBNEUR) * firings.array().transpose().replicate(NBNEUR, 1) * (vneg.array() - THETAVNEG).cwiseMax(0).replicate(1, NBNEUR)
              + ALTP * xplast.array().transpose().replicate(NBNEUR, 1) * (vpos.array() - THETAVNEG).cwiseMax(0).replicate(1, NBNEUR) * (v.array() - THETAVPOS).cwiseMax(0).replicate(1, NBNEUR));
              */
/* // hard mean
vlongtrace *= (double)NBSTEPSVLONGTRACE; vlongtrace -= vnegvals.col(numstep % NBSTEPSVLONGTRACE);
vnegvals.col(numstep % NBSTEPSVLONGTRACE) = (v.array() - THETAVPOS).cwiseMax(0);
vlongtrace += vnegvals.col(numstep % NBSTEPSVLONGTRACE); vlongtrace /= (double)NBSTEPSVLONGTRACE; */

/*
        if ((numstep > 201000) && (firings.sum() > 0)) { toc = clock(); cout << endl << "Time spent, in seconds (excluding stimulus preparation): " << (double)(toc - tic) / CLOCKS_PER_SEC << endl; 
            
            VectorXd LTP1 =      ALTP *  (vpos.array() - THETAVNEG).cwiseMax(0).matrix().cwiseProduct( (v.array() - THETAVPOS).cwiseMax(0).matrix() ) ;
            VectorXd LTD1 =   (-ALTD / VREF2) * vlongtrace.cwiseProduct(vlongtrace).cwiseProduct( (vneg.array() - THETAVNEG).cwiseMax(0).matrix() ) ;  
            for (int nn=0; nn < NBNEUR; nn++)
                EachNeurLTD(nn) =  (-ALTD / VREF2) * vlongtrace(nn) * vlongtrace(nn) * (vneg(nn) - THETAVNEG <0 ? 0 : vneg(nn) -THETAVNEG);
            for (int nn=0; nn < NBNEUR; nn++)
                EachNeurLTP(nn) = ALTP  * (vpos(nn) - THETAVNEG<0 ? 0 : vpos(nn) - THETAVNEG ) * (v(nn) - THETAVPOS < 0 ? 0 : v(nn) - THETAVPOS);
            cout << LTP1.transpose() << endl;
            cout << EachNeurLTP.transpose() << endl << endl << endl;
            cout << LTD1.transpose() << endl;
            cout << EachNeurLTD.transpose() << endl;
            cout << LTP1.maxCoeff() << " :::  " << (LTP1 - EachNeurLTP).cwiseAbs().maxCoeff() << endl;
            
            return 0;}
*/

/*
 
   load resps.txt; resp = resps;
    corrsrRP = corr(resp); % Notice the transpose - columns are image presentations
    %corrsrRP(isnan(corrsrRP(:))) = 0;
    lnk = linkage(double(corrsrRP), 'single', 'correlation');
    figure; [h t oRP] = dendrogram(lnk,size(corrsrRP,1));
    sorted = corrsrRP(oRP',oRP');
    figcorrpres = figure;
    subplot(5,1 , 1:4);
    imagesc(sorted)
    title('Correlations between response patterns');
    subplot(5,1,5)
    sr = sum(resp); % Total firing rates for each pattern presentation
    plot(sr(oRP)); % Ordered in the same pattern order as the matrix
    xlabel('Firing Rates');
    %print(gcf, '-dpng', '-r0', ['clusters_presentations_' FILENAME '.png'])   
    
 */

/*
            firings = (currentlyfiring.array() > -1).select(ZeroV, firings);
            currentlyfiring.array() -= 1;
            currentlyfiring = currentlyfiring.cwiseMax(-1);
            currentlyfiring.array() += firings.array() * 2;
            v = (currentlyfiring.array() > 0).select(VPEAK, v);
            v = (currentlyfiring.array() == 0).select(VRESET, v);
            */
